from unicodedata import numeric
import pandas as pd
import typing as t
from datetime import datetime
from enum import Enum
from tqdm import tqdm
from tables import Table, DATA_TYPES, TABLES
import subprocess
import os
import argparse

from main import MySQLClient

DIR_DBGEN = '/opt/mysql/master_thesis/tpc-h/dbgen/'
DIR_MYSQLD = '/home/aiversen/cs-midt-mt/mysql-server/build/runtime_output_directory/'
DIR_SCRIPT = '/opt/mysql/master_thesis/tpc-h/script/'
DIR_DATA = '/opt/mysql/master_thesis/tpc-h/ref_data_2/'


def convert_table(table: Table, output_dir: str, input_prefix: str):
    for df in pd.read_csv(f"{input_prefix}{table.filename}", sep="|", index_col=False, header=None, chunksize=500000, iterator=True):
        df = df.iloc[: , :-1]
        df.columns = [ col.name for col in table.headers]


        for col in table.headers:
            if col.type == DATA_TYPES.DECIMAL and col.to == DATA_TYPES.INTEGER:
                df[col.name] = 10000 * df[col.name].astype(int)
            
            if col.type == DATA_TYPES.DATE and col.to == DATA_TYPES.INTEGER:
                df[col.name] = df[col.name].map(lambda row: datetime.fromisoformat(row).timestamp() if row != None else None)
                df[col.name] = df[col.name].astype(int)
        
        
        os.makedirs(output_dir, exist_ok=True)
        output_path =  f"{output_dir}/{table.filename}"
        df.to_csv(output_path, mode='a', header=False, index=False, sep="|")


def generate_scale_data(scale: float):
    cwd = DIR_DBGEN
    command = f"./dbgen -s {scale} -f"
    log(f"Running command: {command}")
    subprocess.run(args=command, cwd=cwd, shell=True, text=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

def sync_mysql(scale: float):
    scale_str = str(int(scale)) if int(scale) > 0 else '{:0>2}'.format(int(scale * 10))

    client = MySQLClient(DIR_MYSQLD, f'm_tpch{scale_str}', 0)
   
    client.query(f'SET GLOBAL local_infile=1;', [])

    # Dropping table
    log("Dropping and recreating database")
    client.query(f'DROP DATABASE IF EXISTS {client.dbname}', [])
    client.query(f'CREATE DATABASE IF NOT EXISTS {client.dbname}', [])

    # Setup queries
    log("Running setup")
    client.query(f'SOURCE {DIR_SCRIPT}setup.sql', ['-D', client.dbname, '--local-infile=1'])

    # Insert data
    dir_path = f'{DIR_DATA}{str(scale)}/'

    log("Inserting REGION")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}region.tbl\' INTO TABLE REGION FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting PART")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}part.tbl\' INTO TABLE PART FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting NATION")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}nation.tbl\' INTO TABLE NATION FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting SUPPLIER")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}supplier.tbl\' INTO TABLE SUPPLIER FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting PARTSUPP")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}partsupp.tbl\' INTO TABLE PARTSUPP FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting CUSTOMER")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}customer.tbl\' INTO TABLE CUSTOMER FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting ORDERS")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}orders.tbl\' INTO TABLE ORDERS FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])
    log("Inserting LINEITEM")
    client.query(f'LOAD DATA LOCAL INFILE \'{dir_path}lineitem.tbl\' INTO TABLE LINEITEM FIELDS TERMINATED BY \'|\';', ['-D', client.dbname, '--local-infile=1'])

    
    log("Adding constraints")
    client.query(f"SOURCE {DIR_SCRIPT}constraints.sql", ['-D', client.dbname, '--local-infile=1'])


def log(message):
    print(f"[{datetime.now().strftime('%Y/%m/%d, %H:%M:%S')}]: {message}")

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('--scales', metavar='scales', type=float, help='scales', nargs='+', default=[0.1,1,2,3])
    parser.add_argument('--actions', metavar='actions', type=str, help='generate,convert,insert', nargs='+', default=['generate', 'convert', 'insert']) 
    args = parser.parse_args()

    log(args)

    for scale in args.scales:
        log(f"TPC-H Scale {scale}")
        log("-----------------------------------------")
        if 'generate' in args.actions:
            log("Generating data")
            generate_scale_data(scale)

        if 'convert' in args.actions:
            log(f"Converting tables")
            for table in tqdm(TABLES):
                convert_table(table, output_dir=f"/opt/mysql/master_thesis/tpc-h/ref_data_2/{scale}", input_prefix="/opt/mysql/master_thesis/tpc-h/dbgen/")

        if 'insert' in args.actions:
            log(f"Inserting into MySQL DB")
            sync_mysql(scale)
        log('\n')
