from io import TextIOWrapper
from typing import List, Tuple
import pandas as pd
from tables import DATA_TYPES, TABLE_LINEITEM, Column
import random
from numpy.random import RandomState

random.seed(42069)

class TYPES:
    INT = 0
    STRING = 1
    MIXED = 2

HeaderAndData = Tuple[List[Column], pd.DataFrame]

USED_WHERE_OR = {}

OPERATORS = {
    "L_SHIPDATE": ['<', '>='],
    "L_COMMITDATE": ['<', '>='],
    "L_RECEIPTDATE": ['<', '>='],
    "L_QUANTITY": ['<']
}

def extract_headers_and_data(path: str, columns: List[Column]) -> HeaderAndData:
    headers = [col.name for col in columns]
    df = pd.read_csv(path, delimiter='|', header=None)
    df.columns = headers
    print(df.head())
    return (columns, df)

def filter_headers_and_data_by_type(hd: HeaderAndData, type: TYPES) -> HeaderAndData:
    columns = hd[0]
    df = hd[1]
    if type == TYPES.INT:
        columns = [col for col in columns if col.type == DATA_TYPES.INTEGER or col.to == DATA_TYPES.INTEGER]
    elif type == TYPES.STRING:
        columns = [col for col in columns if col.type in [DATA_TYPES.VARCHAR, DATA_TYPES.CHAR] or col.to in [DATA_TYPES.VARCHAR, DATA_TYPES.CHAR]]
    
    if type != TYPES.MIXED:
        headers = [col.name for col in columns]
        df = df[headers]
    return (columns, df)


def get_random_field_and_value(hd: HeaderAndData, rand_gen: RandomState) -> Tuple[str, str]:
    columns = hd[0]
    df = hd[1]

    rand_col = None
    value = None
    while True:

        rand_col = columns[rand_gen.randint(0, len(columns))]
        value = df.sample(random_state=rand_gen)[rand_col.name].values[0]
        if rand_col.type in [DATA_TYPES.VARCHAR, DATA_TYPES.CHAR]:
            value = f'\"{value}__\"'
        else:
            value = value + 1

        if rand_col.name not in USED_WHERE_OR:
            USED_WHERE_OR[rand_col.name] = {}
            break
        
        if value not in USED_WHERE_OR[rand_col.name]:
            break

    USED_WHERE_OR[rand_col.name][value] = 1
    return (rand_col.name, value)

def gen_or(f: TextIOWrapper, hd: HeaderAndData, rand_gen: RandomState):
    (column, value) = get_random_field_and_value(hd, rand_gen)

    op = "="
    if False and column in OPERATORS:
        # op = OPERATORS[column][random.randint(0, len(OPERATORS[column]) - 1)]
        op = OPERATORS[column][rand_gen.randint(0, len(OPERATORS[column]))]

    f.write(f'{column} {op} {value}')

def gen_or_block(f: TextIOWrapper, hd: HeaderAndData, n: int, rand_gen: RandomState):
    f.write('(\n')
    for i in range(0, n):
        f.write('\t')
        gen_or(f, hd, rand_gen)
        if i < n - 1:
            f.write(f' or\n')
        else:
            f.write('\n')
    f.write(')')

def gen_where(f: TextIOWrapper, hd: HeaderAndData, num_and: int, num_or: int):
    rand_gen = RandomState(6921)
    for i in range(0, num_and):
        gen_or_block(f, hd, num_or, rand_gen)
        if i < num_and - 1:
            f.write('\nand\n')

def gen_query(f: TextIOWrapper, hd: HeaderAndData, num_and: int, num_or: int):
    f.write('select\nsum(l_extendedprice * l_discount) as revenue\nfrom\nLINEITEM\nwhere\n')
    gen_where(f, hd, num_and, num_or)
    pass

if __name__ == '__main__':
    hd = extract_headers_and_data('../ref_data_2/0.1/lineitem.tbl', TABLE_LINEITEM.headers)

    # 1, 10, 25, 50, 100, 250, 500, 1000
    for size in [2000, 4000]:
        for type in [TYPES.INT, TYPES.STRING, TYPES.MIXED]:
            USED_WHERE_OR = {}

            filtered_hd = filter_headers_and_data_by_type(hd, type)
            type_name = 'int' if type == TYPES.INT else 'str' if type == TYPES.STRING else 'mix'
            print(f"Generating type = {type_name} and size = {size}")


            with open(f'../queries_where/query_{size}_{type_name}.sql', 'w') as f:
                gen_query(f, filtered_hd, 1, size)
