#!/usr/bin/python

from enum import Enum
import typing
from typing import Tuple, ByteString
import argparse
import subprocess
from pathlib import PurePath, Path
import time
import os
import sys
from datetime import datetime
import utils
import json
from tables import TABLES

def log(message: str, **kwargs):
    print(f"[{datetime.now()}]: {message}", **kwargs)
class Client():

    def process(self, params: list = [], **kwargs) -> subprocess.Popen:
        pass

    def query(self, query: str, params: list = None, verbose: bool = False) -> typing.Tuple[ByteString, ByteString, int]:
        pass

    def exec_file(self, file: PurePath) -> typing.Tuple[ByteString, ByteString, int]:
        pass

    def enable_jit_timing(self, enable: int):
        pass

    def enable_jit_optimize(self):
        pass

    def enable_jit(self):
        pass

    def enable_instruction_count(self):
        pass

ClientType = typing.NewType('ClientType', Client)

class DB(Enum):
    MYSQL = 1
    POSTGRES = 2
            
class MySQLClient(Client):
    bin_path: PurePath
    dbname: str
    jit: int
    optimize: int

    def __init__(self, path: str, dbname: str, jit: int, optimize: int):
        self.bin_path = PurePath(path, './mysql')
        self.dbname = dbname
        self.jit = jit
        self.optimize = optimize

        self.query(f'CREATE DATABASE IF NOT EXISTS {self.dbname};', [])
        self.enable_hypergraph()
        self.enable_jit()
        self.enable_jit_optimize()

    def process(self, params: list = [], **kwargs) -> subprocess.Popen:
        return subprocess.Popen(['sudo', str(self.bin_path), *params], stdout=subprocess.PIPE, **kwargs)

    def query(self, query: str, params: list = None, verbose: bool = False) -> typing.Tuple[ByteString, ByteString, int]:
        proc = self.process(['-D', self.dbname] if params is None else params, stdin=subprocess.PIPE)
        start = time.time()
        (out, err) = proc.communicate(query.encode('utf-8'))
        end = time.time()
        if verbose and out:
            log(out)
        if verbose and err:
            log(err)

        if proc.returncode == 1:
            log('Query failed! Exiting!')
            sys.exit(1)

        return (out, err, end-start)

    def exec_file(self, file: PurePath) -> typing.Tuple[ByteString, ByteString]:
        with open(str(file), 'r') as f: 
            proc = self.process(['-D', self.dbname], stdin=f)
            start = time.time()
            (out, err) = proc.communicate()
            end = time.time()
            # if out:
            #     log(out)
            # if err:
            #     log(err)
            return (out, err, end-start)

    def enable_jit_timing(self, enable: int):
        return self.query(f'SET @@GLOBAL.should_time_compile_time={enable}; SET @@GLOBAL.should_time_val_int_call={enable};')

    def enable_jit(self):
        return self.query(f'SET @@GLOBAL.should_jit_compile={self.jit};')

    def enable_hypergraph(self):
        return self.query(f"SET GLOBAL optimizer_switch='hypergraph_optimizer=on'")

    def enable_jit_optimize(self):
        return self.query(f'SET @@GLOBAL.should_jit_optimize={self.optimize};')

    def enable_instruction_count(self, enable: int):
        return self.query(f"SET @@GLOBAL.should_count_instructions={enable}")

class PostgresClient(Client):
    bin_path: PurePath
    dbname: str
    jit: int
    parallel: bool

    def __init__(self, path: str, dbname: str, jit: int, parallel: bool):
        self.bin_path = PurePath(path, 'psql')
        self.dbname = dbname
        self.jit = jit
        self.parallel = parallel

        log(f"Jit: {self.jit}")
        log(f"Parallel: {self.parallel}")

        if self.jit < 0 or self.jit > 3:
            log("Jit: invalid value. Should be 0, 1, or 2")
            os.exit(0)

        self.query(f"SELECT 'CREATE DATABASE {self.dbname}' WHERE NOT EXISTS (SELECT FROM pg_database WHERE datname = '{self.dbname}')\gexec;", [])

        self.query(f"ALTER SYSTEM SET jit = {'on' if self.jit != 0 else 'off'}; ALTER SYSTEM SET max_parallel_workers_per_gather={2 if self.parallel else 0}; SELECT pg_reload_conf();")
        if self.jit != 0:
            # Enable jit configuration
            jit_above_cost = 100000 if self.jit == 1 else 0
            jit_inline_above_cost = 500000 if self.jit == 1 else 1000000000 if self.jit == 3 else 0
            jit_optimize_above_cost = 500000 if self.jit == 1 else 10000000000 if self.jit == 3 else 0
            self.query(f"ALTER SYSTEM SET jit_above_cost = {jit_above_cost}; ALTER SYSTEM SET jit_inline_above_cost={jit_inline_above_cost}; ALTER SYSTEM SET jit_optimize_above_cost = {jit_optimize_above_cost}; SELECT pg_reload_conf();")


    def process(self, params: list = [], **kwargs) -> subprocess.Popen:
        return subprocess.Popen(['sudo', str(self.bin_path), *params], stdout=subprocess.PIPE, **kwargs)

    def query(self, query: str, params: list = None, verbose: bool = False) -> typing.Tuple[ByteString, ByteString, int]:
        proc = self.process(['-d', self.dbname] if params is None else params, stdin=subprocess.PIPE)
        start = time.time()
        (out, err) = proc.communicate(query.encode('utf-8'))
        end = time.time()
        if verbose and out:
            log(out)
        if verbose and err:
            log(err)
        return (out, err, end-start)

    def exec_file(self, file: PurePath) -> typing.Tuple[ByteString, ByteString]:
        with open(str(file), 'r') as f: 
            proc = self.process(['-d', self.dbname], stdin=f)
            start = time.time()
            (out, err) = proc.communicate()
            end = time.time()
            if out:
                log(out)
            if err:
                log(err)
            return (out, err, end-start)


class Context:
    client: ClientType
    db_type: DB
    db_name: str
    query_path: str
    client_path: str
    name: str
    verbose: bool
    jit: int
    optimize: int
    parallel: bool
    count_instructions: int

    def __init__(self, db_type: DB, query_path: str, dbname: str, client_path: str, verbose: bool, name: str, jit: int, optimize: int, parallel: bool):
        self.db_type = db_type
        self.query_path = query_path
        self.client_path = client_path
        self.db_name = dbname
        self.name = name if name else str(int(time.time()))
        self.verbose = verbose
        self.jit = jit
        self.optimize = optimize
        self.parallel = parallel

        if(self.db_type == DB.MYSQL):
            self.client = MySQLClient(self.client_path, self.db_name, self.jit, self.optimize)
        elif(self.db_type == DB.POSTGRES):
            self.client = PostgresClient(self.client_path, self.db_name, self.jit, self.parallel)
        
        # for table in TABLES:
        #     log(f"Running 'ANALYZE TABLE {table.name}'")
        #     self.client.query(f"ANALYZE TABLE {table.name};")
            
    def set_mode(self, mode: str):
        if mode == "jit":
            self.jit = 1
            self.optimize = 0
            self.client.jit = 1
            self.client.optimize = 0
        elif mode == "njit":
            self.jit = 0
            self.optimize = 0
            self.client.jit = 0
            self.client.optimize = 0
        elif mode == "jitopt":
            self.jit = 1
            self.optimize = 1
            self.client.jit = 1
            self.client.optimize = 1
        else:
            log(f"Invalid mode: {mode}")
            sys.exit(1)
        
        self.client.enable_jit()
        self.client.enable_jit_optimize()

    @property
    def mode_name(self):
        return f"{self.db_name}_{self.name}_{mode}"

    def analyze_query(self, n: int, iterations: int, count_instruction: bool = False):
        self.client.enable_jit_timing(1)       
        self.client.enable_instruction_count(count_instruction)

        # Warm up
        print()
        for i in range(0, 5):
            log(f"\tRunning warm up {i+1}/5", end='\r')
            _ = self.exec_query(n, 'EXPLAIN ANALYZE ', '')
        print()

        query_plans = {}

        res = None
        explain_data = utils.ExplainAnalyzeData()
        for i in range(0, iterations):
            log(f"\tRunning iteration {i+1}/{iterations}", end='\r')
            cur_res = self.exec_query(n, 'EXPLAIN ANALYZE ', '')
            data = utils.extract_time(utils.split_newline(cur_res[0]))
            data.set_total_time(cur_res[2])

            query_plan_str = ':'.join([d["name"] for d in data.raw_data if d["name"] is not None])
            if query_plan_str not in query_plans:
                query_plans[query_plan_str] = [i]
            else:
                query_plans[query_plan_str].append(i)

            explain_data.add(data)
            res = cur_res
        print()

        if len(query_plans.keys()) > 1:
            log(f"{utils.COLORS.FAIL}Received different query plans in analyze query!{utils.COLORS.ENDC}")
            for key, value in query_plans.items():
                log(f"\t{key}: {value}")

        explain_data.divide(iterations)
        log(f"\tval_int_calls = {explain_data.val_int_call_time}s")

        if res is not None:
            # Write end result to file
            self.write_result(f'{n}.analyze.sql', res)

        self.write_result(f'{n}.analyze.metadata.sql', ({
            "val_int_call_time": explain_data.val_int_call_time,
            "codegen": explain_data.codegen_time,
            "compile": explain_data.compile_time,
            "total_time": explain_data.total_time,
            "opt": explain_data.optimization_time,
            "instruction_count": explain_data.instruction_count,
        },))

        return res

    def benchmark_query(self, n: int, iterations: int, save_results: bool = False):
        self.client.enable_jit_timing(0)
        self.client.enable_instruction_count(0)     

        times = []
        for i in range(0, iterations):
            log(f"\tRunning iteration {i+1}/{iterations}", end='\r')
            res = self.exec_query(n, '', '')
            times.append(res[2])
            if i == 0 and save_results:
                self.write_result(f'{n}.results.sql', (res[0], res[1]))
        print()
        res = (sum(times)/len(times), min(times), max(times))
        self.write_result(f'{n}.benchmark.sql', res)
        return res

    def analyze_where_query(self, query_file: PurePath, iterations: int, count_instruction: bool = False):
        query = read_query_from_file(query_file)
        self.client.enable_jit_timing(1)       
        self.client.enable_instruction_count(count_instruction)       

        components = query_file.name.strip(".sql").split("_")
        size, type = components[1], components[2]

        res = None
        explain_data = utils.ExplainAnalyzeData()
        for i in range(0, iterations):
            log(f"\tRunning iteration {i+1}/{iterations}", end='\r')
            res = self.client.query(f'EXPLAIN ANALYZE {query};', verbose=self.verbose)
            data = utils.extract_time(utils.split_newline(res[0]))
            data.set_total_time(res[2])
            explain_data.add(data)
        explain_data.divide(iterations)
        print(explain_data)
        log(f"\tval_int_calls = {explain_data.val_int_call_time}s")

        if res is not None:
            # Write end result to file
            self.write_result(f'where.analyze.{type}.{size}.sql', res)

        self.write_result(f'where.analyzedata.{type}.{size}.sql', ({
            "val_int_call_time": explain_data.val_int_call_time,
            "codegen": explain_data.codegen_time,
            "compile": explain_data.compile_time,
            "total_time": explain_data.total_time,
            "opt": explain_data.optimization_time,
            "instruction_count": explain_data.instruction_count,
        },))

        return res

    def benchmark_where_query(self, query_file: PurePath, iterations: int, save_results: bool = False):
        query = read_query_from_file(query_file)
        self.client.enable_jit_timing(0)              

        components = query_file.name.strip(".sql").split("_")
        size, type = components[1], components[2]

        times = []
        for i in range(0, iterations):
            log(f"\tRunning iteration {i+1}/{iterations}", end='\r')
            res = self.client.query(f'{query};', verbose=self.verbose)
            times.append(res[2])
            if i == 0 and save_results:
                self.write_result(f'where.benchmark.results.{type}.{size}.sql', (res[0], res[1]))
        print()
        res = (sum(times)/len(times), min(times), max(times))
        self.write_result(f'where.benchmark.{type}.{size}.sql', res)
        return res

    def exec_query(self, n: int, prefix: str, suffix: str):
        path = PurePath(self.query_path + f'/{n}.sql')

        query = read_query_from_file(path)
        query = prefix + query + suffix

        res = self.client.query(query, verbose=self.verbose)
        return res

    def exec_file(self, file: str):
        self.client.exec_file(PurePath(file))

    def write_result(self, name: str, res: Tuple[ByteString, ByteString]):
        path = Path(f'data', 'mysql' if self.db_type == DB.MYSQL else 'pg', f'{self.mode_name}')
        path.mkdir(parents=True, exist_ok=True)

        path = Path(path, name)
        with open(str(path), 'w') as f:
            for byte_str in res:
                if byte_str is not None:
                    if isinstance(byte_str, bytes):
                        f.write(byte_str.decode('utf-8'))
                    elif isinstance(byte_str, dict):
                        f.write(json.dumps(byte_str))
                    else:
                        f.write(str(byte_str))
                    f.write('\n')


def read_query_from_file(path: PurePath) -> str:
    query: str = ''
    with open(str(path)) as f:
        query = f.read()
    return query


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('jobs', metavar='jobs', type=str, default='queries', help='"init,analyze,benchmark,viz"')
    parser.add_argument('modes', metavar='modes', type=str, default='jit,jitopt,njit', help='"jit,jitopt,njit"')
    parser.add_argument('--dbtype', metavar='dbtype', type=str, default='MYSQL', help='the database')
    parser.add_argument('--query_path', '--qp', metavar='qp', type=str, default='./', help='query path')
    parser.add_argument('--db_name', metavar='dbname', type=str, default='tpch', help='the database name')
    parser.add_argument('--client_path', '--cp', metavar='cp', type=str, default='', help='the path to the client')
    parser.add_argument('--verbose', '--v', metavar='verbose', type=bool, default=False, help='the path to the client')
    parser.add_argument('--n', '--n', metavar='n', type=int, default=10, help='number of iterations in benchmark')
    parser.add_argument('--analyze_n', metavar='analyze_n', type=int, default=10, help='number of iterations in analyze')
    parser.add_argument('--warmup_n', metavar='warmup_n', type=int, default=4, help='number of warmup iterations')
    parser.add_argument('--name', metavar='name', type=str, default='', help='the name of the run')
    parser.add_argument('--exclude', metavar='exclude', type=int, help='queries to exclude', nargs='+')
    parser.add_argument('--include', metavar='include', type=int, help='queries to include', nargs='+')
    parser.add_argument('--parallel', metavar='parallel', type=bool, default=False, help='enable parallell workers')
    parser.add_argument('--count_instructions', metavar='count_instructions', type=bool, default=False, help='should count instructions')
    parser.add_argument('--save-results', metavar='save_results', type=bool, default=False, help='save results')
    args = parser.parse_args()

    dbtype = DB.MYSQL if args.dbtype == 'MYSQL' else DB.POSTGRES 

    if args.exclude is not None and args.include is not None and len(args.exclude) > 0 and len(args.include) > 0:
        log(f"Invalid arguments: both 'include' and 'exclude' can not be specified")
        os.exit(1)

    print(dbtype)
    if args.exclude and len(args.exclude) > 0:
        print(f"Exclude: {args.exclude}")
    elif args.include and len(args.include) > 0:
        print(f"Queries: {args.include}")
    else:
        print("Queries: all")
    print(f"Database: {args.db_name}")
    print(f"Name: {args.name}")

    ctx = Context(dbtype, args.query_path, args.db_name, args.client_path, args.verbose, args.name, 0, 0, args.parallel)

    

    if "init" in args.jobs:
        log("===Initializing===")
        # TODO: Fix running build_db.sql scripts

    if "analyze" in args.jobs:
        log(f"{utils.COLORS.LIGHT_BLUE}===Running EXPLAIN ANALYZE on queries==={utils.COLORS.ENDC}")
        log(f"Will count instructions: {args.count_instructions}")
        for i in range(1, 23):
            if args.exclude and i in args.exclude:
                continue
            if args.include and i not in args.include:
                continue
            log(f"Q{i}: ")
            for mode in args.modes.split(','):
                log(f"===========Mode: {mode}=============")
                ctx.set_mode(mode)

                res = ctx.analyze_query(i, args.analyze_n, args.count_instructions)
                log(f"\tQuery time: {res[2]}s")

    
    if "benchmark" in args.jobs:
        log(f"{utils.COLORS.LIGHT_BLUE}===Running BENCHMARK(n = {args.n}) on queries==={utils.COLORS.ENDC}")
        for i in range(1, 23):
            if args.exclude and i in args.exclude:
                continue
            if args.include and i not in args.include:
                continue
            log(f"Q{i}:")
            for k in range(0, args.warmup_n):
                log(f"\tRunning warm up {k+1}/{args.warmup_n}", end='\r')
                _ = ctx.exec_query(i, '', '')
            print()
            for mode in args.modes.split(','):
                log(f"===========Mode: {mode}=============")
                ctx.set_mode(mode)

                res = ctx.benchmark_query(i, args.n, args.save_results)
                log(f"\tavg: {res[0]}")
                log(f"\tmin: {res[1]}")
                log(f"\tmax: {res[2]}")

    if "where" in args.jobs:
        log(f"{utils.COLORS.PURPLE}===Running WHERE(n = {args.n}) on queries==={utils.COLORS.ENDC}")
        log(f"Will count instructions: {args.count_instructions}")
        for file in os.listdir(args.query_path):
            if file.endswith(".sql"):
                log("")
                log(f"{file}:")
                for k in range(0, args.warmup_n):
                    log(f"\tRunning warm up {k+1}/{args.warmup_n}", end='\r')
                    _ = ctx.exec_file(str(PurePath(args.query_path, file)))
                log("\t----- RUNNING ANALYZE -----")
                for mode in args.modes.split(','):
                    log(f"===========Mode: {mode}=============")
                    ctx.set_mode(mode)

                    res = ctx.analyze_where_query(PurePath(args.query_path, file), args.n, args.count_instructions)
                    log(f"\tquery_time = {res[2]}s")

                # log("\t----- RUNNING BENCHMARK -----")
                # for mode in args.modes.split(','):
                #     log(f"===========Mode: {mode}=============")
                #     ctx.set_mode(mode)

                #     res = ctx.benchmark_where_query(PurePath(args.query_path, file), args.n)
                #     log(f"\tquery_time = {res[2]}s")

    if "test" in args.jobs:
        for i in range(0, 22):
            res = ctx.analyze_query(2, 2)
            ctx.write_result(f'{i}.analyze.sql', res)
        
    if "viz" in args.jobs:
        log("===Viz===")
